package org.w3c.rdf.util;

import java.util.*;
import org.w3c.rdf.model.*;
import org.w3c.rdf.vocabulary.rdf_syntax_19990222.RDF;
import org.w3c.rdf.vocabulary.rdf_schema_19990303.RDFS;

/**
 * Limitation: does not work properly if ground model changes. Need ObservableModel for that.
 */

public class OrderedModel implements VirtualModel {

  Model peer;
  Hashtable containers;
  boolean cache = true; // by default cache Containers

  public OrderedModel(Model peer) {

    this.peer = peer;
  }

  public OrderedModel(Model peer, boolean cache) {

    this(peer);
    this.cache = cache;
  }

  private void findContainers() throws ModelException {

    if(cache && containers == null) {

      Model c = peer.find(null, RDF.type, RDF.Bag);
      SetOperations.unite(c, peer.find(null, RDF.type, RDF.Seq));
      SetOperations.unite(c, peer.find(null, RDF.type, RDF.Alt));
                        
      containers = new Hashtable();
                        
      for(Enumeration en = c.elements(); en.hasMoreElements();) {
        Resource r = ((Statement)en.nextElement()).subject();
        containers.put(r, r);
      }

      //      System.err.println("Found " + containers.size() + " containers");
    }
  }

  private boolean isContainer(RDFNode r) throws ModelException {

    if(containers != null)
      return containers.contains(r);
    else
      return findIsContainer(r);
  }

  private boolean findIsContainer(RDFNode n) throws ModelException {

    if(!(n instanceof Resource))
      return false;
    Resource r = (Resource)n;

    return (!peer.find(r, RDF.type, RDF.Seq).isEmpty()) ||
      (!peer.find(r, RDF.type, RDF.Bag).isEmpty()) ||
      (!peer.find(r, RDF.type, RDF.Alt).isEmpty());
  }

  private boolean affectsContainer(Statement s) throws ModelException {

    //    return findIsContainer(s.object()) || findIsContainer(s.subject());
    return
      RDF.Bag.equals(s.object()) ||
      RDF.Seq.equals(s.object());
  }

  public Model getGroundModel() {

    return peer;
  }

  public String getLabel() throws ModelException {

    return peer.getLabel();
  }

  public String getURI() throws ModelException {

    return peer.getURI();
  }

  /**
   * Set a base URI for the model.
   * Affects creating of new resources and serialization syntax.
   * Inherited method getURI returns the URI set in this method
   */
  public void setSourceURI(String uri) throws ModelException {
    peer.setSourceURI(uri);
  }

  /**
   * Returns current base URI setting.
   */
  public String getSourceURI() throws ModelException {
    return peer.getSourceURI();
  }

  // Model access

  /**
   * Number of triples in the model
   *
   * @return  number of triples, -1 if unknown
   * 
   * @seeAlso org.w3c.rdf.model.VirtualModel
   */
  public int size() throws ModelException {

    // FIXME: not quite faithful
    return find(null, null, null).size();
  }

  /**
   * true if the model contains no triples
   */
  public boolean isEmpty() throws ModelException {

    return peer.isEmpty();
  }

  /**
   * Enumerate triples
   */
  public Enumeration elements() throws ModelException {

    return find(null, null, null).elements();
  }

  /**
   * Tests if the model contains the given triple.
   *
   * @return  <code>true</code> if the triple belongs to the model;
   *          <code>false</code> otherwise.
   */
  public boolean contains(Statement t) throws ModelException {

    // FIXME: efficiency?
    return !find(t.subject(), t.predicate(), t.object()).isEmpty();
  }


  // Model manipulation: add, remove, find

  /**
   * Adds a new triple to the model.
   */
  public void add(Statement t) throws ModelException {

    if(cache && affectsContainer(t)) {
      containers = null;
    }
    peer.add(t);
  }

  /**
   * Removes the triple from the model.
   */
  public void remove(Statement t) throws ModelException {

    if(cache && affectsContainer(t)) {
      containers = null;
    }
    peer.remove(t);
  }

  /**
   * General method to search for triples.
   * <code>null</code> input for any parameter will match anything.
   * <p>Example: <code>Model result = m.find( null, RDF.type, new ResourceImpl("http://...#MyClass") )</code>
   * <p>finds all instances of the class <code>MyClass</code>
   */
  public Model find( Resource subject, Resource predicate, RDFNode object ) throws ModelException {

    findContainers();

    Model m = peer.find(subject, predicate, object);

    Model derived = null, subtract = null;
    NodeFactory f = m.getNodeFactory();

    for(Enumeration en = m.elements(); en.hasMoreElements();) {

      Statement st = (Statement)en.nextElement();

      //                        System.err.println("Examining " + st);

      Resource s = st.subject();
      Resource p = st.predicate();
      RDFNode o = st.object();
      boolean co = isContainer(o);
      boolean cs = isContainer(s);

      if(co || cs) {
        if(subtract == null)
          subtract = peer.create();
        subtract.add(st);
      }

      if(cs && (RDFUtil.getOrd(p) > 0 || RDF.type.equals(p))) {
                                // ignore _n and type arcs
        continue; 
      }

      if(derived == null && (cs || co)) {
        derived = peer.create();
      }
      if(cs) { // subject is a Bag

        int i = 0;
        RDFNode r1 = null;
        do {
          r1 = RDFUtil.getObject(peer, s, f.createOrdinal(++i));

          if(r1 instanceof Literal) // cannot express that...
            continue;
          //                                    System.err.println(f.createOrdinal(i) + " of " + s + " is " + r1);

          if(r1 != null) {

            if(co) { // object is a Bag too
              int j = 0;
              Resource r2 = null;
              do {
                r2 = RDFUtil.getObjectResource(peer, (Resource)o, f.createOrdinal(++j));
                if(r2 != null) {
		  //		  System.out.println("DERIVED1: " + f.createStatement((Resource)r1, p, r2));
                  derived.add(f.createStatement((Resource)r1, p, r2));
		}
              } while (r2 != null);

            } else { // object is not a Bag
	      //	      System.out.println("DERIVED2: " + f.createStatement((Resource)r1, p, o));
              derived.add(f.createStatement((Resource)r1, p, o));
            }
          }
        } while (r1 != null);

      } else if(co) { // only object is a Bag

        int j = 0;
        RDFNode r2 = null;
        do {
          r2 = RDFUtil.getObject(peer, (Resource)o, f.createOrdinal(++j));
          if(r2 != null)
            derived.add(f.createStatement(s, p, r2));
        } while (r2 != null);
      }
    }

    if(derived != null) {

      System.out.println("DERIVED: ");
      RDFUtil.printStatements(derived, System.out);
      System.out.println();

      SetOperations.unite(m, derived);
    }

    if(subtract != null)
      SetOperations.subtract(m, subtract);

    return m;
  }

  /**
   * Search for a single triple.
   * <code>null</code> input for any parameter will match anything.
   *
   * @return  a single triple if such was found;<br>
   *          <code>null</code> if nothing was found;<br>
   *          otherwise a RuntimeException is thrown.
   public Statement find1( Resource subject, Resource predicate, RDFnode object );
  */

  /**
   * Clone the model.
   */
  public Model duplicate() throws ModelException {

    return new OrderedModel(peer.duplicate());
  }

  /**
   * Creates empty model of the same Class
   */
  public Model create() throws ModelException {

    return new OrderedModel(peer.create());
  }

  public String toString() {

    try {
      return "[OrderedModel " + getURI() + "]";
    } catch (ModelException ex) {
      return "[OrderedModel: " + ex + "]";
    }
  }
  /**
   * Returns the node factory for this model
   */
  public NodeFactory getNodeFactory() throws ModelException {

    return peer.getNodeFactory();
  }

  public boolean isMutable() throws ModelException {

    return peer.isMutable();
  }

  public static void main(String[] args) throws Exception {

    // testing closure
    if(args.length == 0) {
      System.err.println("OrderedModel <fileOrURL>");
      System.exit(1);
    }

    String fileNameOrURL = args[0];
    RDFFactory f = new RDFFactoryImpl();
    Model m = f.createModel();
    RDFUtil.parse(fileNameOrURL, f.createParser(), m);
                
    //    Model om = new OrderedModel(m).find(null, m.getNodeFactory().createResource("http://www.omg.org/uml/1.3/Foundation.Core.Association.connection"), null);
    System.out.println("All statements: ");
    RDFUtil.printStatements(m, System.out);
  }

        
}
